{-# LANGUAGE DoAndIfThenElse
           , GeneralizedNewtypeDeriving
           , OverloadedStrings
           , TypeFamilies
  #-}
{-# OPTIONS_GHC -fno-warn-orphans #-}

module Forest (initialiseForest) where

import           Control.Applicative ((<$>))
import           Control.Arrow ((&&&), first)
import           Control.DeepSeq (NFData, rnf)
import           Control.Monad (liftM, replicateM, zipWithM, zipWithM_)
import           Control.Monad.IO.Class (liftIO)
import           Control.Monad.Trans (lift)
import           Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Char8 as BC
import           Data.Char (isDigit)
import           Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import qualified Data.List as L
import           Data.Map (Map)
import           Data.Monoid (Monoid, mempty, mappend)
import           Data.Tuple (swap)
import           System.Directory (createDirectoryIfMissing)
import           System.FilePath ((</>))
import           System.Random (StdGen, mkStdGen)

import           Combine (Mean(..), combine)
import qualified IntBiMap as IBM
import qualified Interpreter as I
import           KNN (knn)
import           Predictor
import           RandomState
import           Types
import           Utils


data Forest = Forest {
  trees     :: [(Tree, StdGen)]
, rng       :: StdGen
}

data DepthWeightAlgorithm = Ascending | Descending | Inverse | Constant
  deriving (Show, Read)

data PredictionAlgorithm = KNN | Naive
  deriving (Show, Read)

data TreeBuildStopCriterion = Logarithmic | Null | Sqrt
  deriving (Show, Read)

data ConfidenceAlgorithm = NaiveConfidence | ConstConfidence
  deriving (Show, Read)

data SampleStrategy = TreesDrawSamples | SamplesDrawTrees
  deriving (Show, Read)

instance Predictor Forest where
  data PredictorCfg Forest = ForestCfg ForestConfig

  learn      = learnForest
  predict    = predictForest
  saveState  = saveForestGraphs
  loadState  = loadForestGraphs

type ForestM = PredictorM Forest
type ForestT = PredictorT Forest

type RandomForestM = RandomStateT StdGen ForestM


-- -----------------------------------------------------------------------------
-- Configuration

data ForestConfig = ForestConfig {
-- | Number of trees in the forest.
  nTrees         :: Int
-- | Number of times a sample is learned over all trees.
, sampleFreq     :: Int
-- | In function of the number of samples at a branch, stop tree build?
, buildStop      :: TreeBuildStopCriterion
-- | Stop tree building at a certain depth.
, maxTreeDepth   :: Int

-- | How do we select samples for each tree?
, sampleStrat    :: SampleStrategy
-- | Seed for the random number generator.
, seed           :: Int

-- | Number of random features to choose for evaluation.
, randFeats      :: Int
-- | Number of features for which Gini index will be calculated.
, giniFeats      :: Int
-- | Method to calculate confidence of a branch.
, confidenceAlg  :: ConfidenceAlgorithm

-- | Algorithm to determine quality of a tree path.
, depthWeightAlg :: DepthWeightAlgorithm
-- | How strongly errors at lower depths will be punished in inverse curve.
-- Values near 1.0 will punish lower depth errors very little, whereas
-- lower values will punish lower depth errors more.
, depthWeightExp :: Double
-- Minimum score for an error.
, minDepthWeight :: Double

-- | Method to combine leafs and path scores to form a prediction.
, predAlg        :: PredictionAlgorithm
-- | Weight of a label in contrast to the weight of the sample itself.
, depWeight      :: Double
-- | Number of predictions produced by each tree.
, predsPerTree   :: Int
-- | Calculate IDF for samples before running k-NN.
, idf            :: Bool
-- | How much to punish missing features in naive prediction mode.
, labelDeltaExp  :: Double

-- | Mean (arithmetic, ...) to combine predictions from different trees.
, combineMean    :: Mean
} deriving Show

defaultConf :: ForestConfig
defaultConf = ForestConfig {
  nTrees         = 4
, sampleFreq     = 16
, buildStop      = Logarithmic
, maxTreeDepth   = 10000

, sampleStrat    = SamplesDrawTrees
, seed           = 42

, randFeats      = 0
, giniFeats      = 0
, confidenceAlg  = ConstConfidence

, depthWeightAlg = Inverse
, depthWeightExp = 0.8
, minDepthWeight = 0.8

, predAlg        = KNN
, depWeight      = 1.5
, predsPerTree   = 1024
, idf            = True
, labelDeltaExp  = 0.1

, combineMean    = Harmonic
}

confActions :: [(String, ForestConfig -> String -> ForestConfig)]
confActions =
  [ ("nTrees"        , \ c v -> c {nTrees         = read v})
  , ("depthWeightAlg", \ c v -> c {depthWeightAlg = read v})
  , ("depWeight"     , \ c v -> c {depWeight      = read v})
  , ("predAlg"       , \ c v -> c {predAlg        = read v})
  , ("labelDeltaExp" , \ c v -> c {labelDeltaExp  = read v})
  , ("giniFeats"     , \ c v -> c {giniFeats      = read v})
  , ("sampleFreq"    , \ c v -> c {sampleFreq     = read v})
  , ("buildStop"     , \ c v -> c {buildStop      = read v})
  , ("confidenceAlg" , \ c v -> c {confidenceAlg  = read v})
  , ("combineMean"   , \ c v -> c {combineMean    = read v})
  , ("predsPerTree"  , \ c v -> c {predsPerTree   = read v})
  , ("randFeats"     , \ c v -> c {randFeats      = read v})
  , ("sampleStrat"   , \ c v -> c {sampleStrat    = read v})
  , ("maxTreeDepth"  , \ c v -> c {maxTreeDepth   = read v})
  , ("minDepthWeight", \ c v -> c {minDepthWeight = read v})
  , ("depthWeightExp", \ c v -> c {depthWeightExp = read v})
  , ("idf"           , \ c v -> c {idf            = read v})
  , ("seed"          , \ c v -> c {seed           = read v})
  ]

asksFCfg :: (ForestConfig -> a) -> ForestM a
asksFCfg f = askConfig >>= \ (ForestCfg c) -> return $ f c


-- -----------------------------------------------------------------------------
-- Tree

data Tree = Leaf Samples | Branch {
  left, right :: Tree
, feature     :: FeatureID
, confidence  :: Double
}

instance NFData Tree where
  rnf (Leaf smpls) = rnf smpls
  rnf (Branch {left = l, right = r, feature = f, confidence = c}) =
    rnf l `seq` rnf r `seq` rnf f `seq` rnf c

-- Unfortunately, this would create an orphan instance warning. But we are
-- safe, because we do not do anything in the 'rnf' call, so we disable the
-- warning.
instance NFData StdGen where
  rnf _ = ()


emptyTree :: Tree
emptyTree = Leaf []


-- -----------------------------------------------------------------------------
-- Tree flattening

-- | Accumulate the samples among all tree leafs.
treeSamples :: Tree -> Samples
treeSamples (Leaf smpls) = smpls
treeSamples (Branch l r _ _) = treeSamples l ++ treeSamples r

-- | Depth of a node in a tree.
newtype Depth = Depth Int deriving (Num, Ord, Eq, Show)
-- | An error contains the depth and the confidence of the node it appeared at.
newtype Error = Error (Depth, Double)
-- | A path from root to leaf stores leaf depth and all errors along the path.
newtype Path = Path (Depth, [Error])

-- | Given some features, list tree paths and the samples at their leafs.
flattenTree :: Features -> Tree -> [(Path, Samples)]
flattenTree ftrs = flatten (Depth 0) [] where
  flatten depth errors (Leaf smpls) = [(Path (depth, errors), smpls)]
  flatten depth errors (Branch l r f w)
    | f `elem` ftrs = fSD errors l ++ fSD (Error (depth, w):errors) r
    | otherwise     = fSD errors r ++ fSD (Error (depth, w):errors) l
    where fSD = flatten (depth + 1)

flattenWeightSamples :: [(Double, Samples)] -> [(SampleID, Double)]
flattenWeightSamples = concatMap (\ (w, smpls) -> zip smpls $ repeat w)


-- -----------------------------------------------------------------------------
-- Tree querying

type ForestPredictor = Features -> [(Double, Samples)] -> ForestM Labels

-- | Return the currently active prediction algorithm.
predictor :: ForestM ForestPredictor
predictor = asksFCfg predAlg >>= \ pA -> return $ case pA of
  Naive -> predictNaive
  KNN   -> predictKNN

-- | Predict labels for a single tree.
query :: Features -> Tree -> ForestM Labels
query ftrs t = predictor >>= \ p ->
  p ftrs =<< mapM (firstM pathScore) (flattenTree ftrs t)

-- | Determine quality of a tree path (by evaluating errors among path).
pathScore :: Path -> ForestM Double
pathScore (Path (Depth mD, errors)) = product <$> mapM f errors where
  f (Error (Depth d, c)) = do

    dWA <- asksFCfg depthWeightAlg
    dWE <- asksFCfg depthWeightExp
    mDW <- asksFCfg minDepthWeight

    return $ c * case dWA of
      Ascending  -> mDW + curve mDW d
      Descending -> 1 - curve mDW d
      Inverse    -> 1 - (1 - mDW) / (fI d + 1)**dWE
      Constant   -> mDW

  curve mDW d = (1 - mDW) * (fI d / fI mD)
  fI = fromIntegral
  

predictKNN :: ForestPredictor
predictKNN ftrs flat = do
  (lT, fT) <- asksLData (sampleLabels &&& sampleFeatures)
  iDF      <- asksFCfg idf
  pPT      <- asksFCfg predsPerTree
  dW       <- asksFCfg depWeight

  let flat' = flattenWeightSamples flat
      fFreq = featureFreq fT IM.empty $ map fst flat'
      fWeights = if iDF then IM.map (** 6) $ featureIDFs (length flat') fFreq
                 else IM.empty

  return $ knn lT fT pPT dW ftrs fWeights flat'

predictNaive :: ForestPredictor
predictNaive ftrs flat = do
  (lT, fT) <- asksLData (sampleLabels &&& sampleFeatures)
  pPT      <- asksFCfg predsPerTree
  dW       <- asksFCfg depWeight
  lDE      <- asksFCfg labelDeltaExp

  let missingFeats s = filter (`notElem` getSampleFeatures fT s) ftrs
      ftrsN = fromIntegral $ length ftrs
      smplScore s = (fromIntegral (length $ missingFeats s) / ftrsN) ** lDE
      addSample w smpl@(SampleID s) = IM.insertWith (+) s (w * smplScore smpl)
      addSamples (score, smpls) preds = foldr (addSample score) preds smpls

  let preds  = addLabels dW lT $ foldr addSamples IM.empty flat
      listed = map swap $ IM.toList preds
      sorted = L.sortBy (flip compare) listed

  return $ take pPT $ map snd sorted

  where
    -- | Add value to an IntMap at all supplied indices.
    addAtIndices x = flip $ foldr (\ i -> IM.insertWith (+) i x)

    addLabels dW lT m = IM.foldWithKey (addSampleLabels dW lT) m m

    addSampleLabels dW lT smpl score =
       let lbls = getSampleLabels lT (SampleID smpl) in
       addAtIndices (depsScore dW score lbls) lbls

    depsScore dW s deps = dW * s / log (fromIntegral $ length deps + 1)


-- -----------------------------------------------------------------------------
-- Label statistics

-- | Label statistics contain a map from labels to their frequencies,
-- as well as the sum of all frequencies.
data LabelStats = LabelStats LabelFreq Int

instance Monoid LabelStats where
  mempty = LabelStats IM.empty 0
  mappend (LabelStats lF1 t1) (LabelStats lF2 t2) =
    LabelStats (IM.unionWith (+) lF1 lF2) (t1 + t2)

initLabelStats :: Labels -> LabelStats
initLabelStats lbls = LabelStats (foldl addAtKeys IM.empty lbls) (length lbls)
  where addAtKeys m k = IM.insertWith (+) k 1 m

updateLabelStats :: SampleLabels -> LabelStats -> SampleID -> LabelStats
updateLabelStats lT lS smpl = mappend lS $ initLabelStats lbls
  where lbls = labelOfSample smpl : getSampleLabels lT smpl

labelStats :: SampleLabels -> [SampleID] -> LabelStats
labelStats lT = foldl (updateLabelStats lT) mempty

minusIntMap :: Num a => IntMap a -> IntMap a -> IntMap a
m1 `minusIntMap` m2 = IM.mapWithKey (\ k x -> x - IM.findWithDefault 0 k m2) m1

minusLabelStats :: LabelStats -> LabelStats -> LabelStats
LabelStats lF1 n1 `minusLabelStats` LabelStats lF2 n2 =
  LabelStats (lF1 `minusIntMap` lF2) (n1 - n2)


-- -----------------------------------------------------------------------------
-- Feature statistics

-- | Record for each feature which labels appear in conjunction with it.
type FeatureStats = IntMap LabelStats

emptyFeatureStats :: Features -> FeatureStats
emptyFeatureStats ftrs = IM.fromList $ zip ftrs $ repeat mempty

updateFeatureStats
  :: SampleLabels -> SampleFeatures -> FeatureStats -> SampleID -> FeatureStats
updateFeatureStats lT fT fS smpl =
  foldr (IM.adjust (\ lS -> updateLabelStats lT lS smpl)) fS feats
  where feats = getSampleFeatures fT smpl

featureStats
  :: SampleLabels -> SampleFeatures -> Features -> Samples -> FeatureStats
featureStats lT fT = foldl (updateFeatureStats lT fT) . emptyFeatureStats


-- -----------------------------------------------------------------------------
-- Tree splitting criteria

giniIndex :: LabelStats -> Double
giniIndex (LabelStats _  0) = 0
giniIndex (LabelStats lF t) = sum $ map (\ p -> p * (1 - p)) ps
  where ps = map (\ (_, n) -> fromIntegral n / fromIntegral t) $ IM.toList lF

giniScore :: LabelStats -> LabelStats -> Double
giniScore lStats trueStats = (trueN * trueScore + falseN * falseScore) / (trueN + falseN)
  where tfStats = (trueStats, lStats `minusLabelStats` trueStats)
        (trueScore, falseScore) = mapPair giniIndex tfStats
        (trueN, falseN) = mapPair (\ (LabelStats _ t) -> fromIntegral t) tfStats

naiveScore :: Double -> Int -> Double
naiveScore smplsH freq = abs $ fromIntegral freq - smplsH


splitCandidates :: Samples -> RandomForestM FeatureFreq
splitCandidates smpls = do
  fT <- lift $ asksLData sampleFeatures
  rF <- lift $ asksFCfg randFeats

  let fF = featureFreq fT IM.empty smpls

  if rF == 0 then return fF
  else IM.fromList <$> sampleWithReplacement rF (IM.toList fF)

splitConfidence :: ConfidenceAlgorithm -> Double -> Double -> Double
splitConfidence ConstConfidence _ _ = 1
splitConfidence NaiveConfidence smplsH s = (1 - s) / smplsH

bestSplit :: Samples -> FeatureFreq -> RandomForestM (FeatureID, Double)
bestSplit smpls candidates = do
  (lT, fT) <- lift $ asksLData (sampleLabels &&& sampleFeatures)
  gF <- lift $ asksFCfg giniFeats
  cA <- lift $ asksFCfg confidenceAlg

  let smplsH = fromIntegral (length smpls) / 2
  let bestFN = sortMapIM (naiveScore smplsH) candidates

  let fStats = featureStats lT fT (map fst $ take gF bestFN) smpls
      lStats = labelStats lT smpls
      bestFG = sortMapIM (giniScore lStats) fStats

  let bestF = fst $ head $ if gF == 0 then bestFN else bestFG
  return (bestF, splitConfidence cA smplsH $ snd $ head bestFN)


-- -----------------------------------------------------------------------------
-- Tree update

updateTree' :: (Tree, StdGen) -> Samples -> ForestM (Tree, StdGen)
updateTree' (t, g) smpls = do
  bS <- asksFCfg buildStop

  runRandomStateT (updateTree (Depth 0) (minSamples bS) t smpls) g

  where minSamples Logarithmic = ceiling $ log nSamples
        minSamples Null = 0
        minSamples Sqrt = ceiling $ sqrt nSamples
        nSamples = fromIntegral $ length smpls :: Double

updateTree :: Depth -> Int -> Tree -> Samples -> RandomForestM Tree
updateTree _ _ t [] = return t
updateTree depth minSamples t smpls = do
  fT <- lift $ asksLData sampleFeatures
  mTD <- lift $ asksFCfg maxTreeDepth

  -- calculate best splitting feature among new and old labels
  let allSmpls = smpls ++ treeSamples t
  (bestF, bestC) <- bestSplit allSmpls =<< splitCandidates allSmpls

  let (oldSamples, yesTree, noTree) = case t of
        Branch l r f _
          | f == bestF -> ([], l, r)
          | otherwise  -> (treeSamples l ++ treeSamples r, emptyTree, emptyTree)
        Leaf lbls -> (lbls, emptyTree, emptyTree)

  let newSamples = smpls ++ oldSamples
  let (yesS, noS) = L.partition (elem bestF . getSampleFeatures fT) newSamples

  if (depth >= Depth mTD) ||
     (length yesS + length (treeSamples yesTree) <= minSamples) ||
     (length  noS + length (treeSamples  noTree) <= minSamples)
  then return $ Leaf allSmpls
  else do
    yesTree' <- updateTree (depth + 1) minSamples yesTree yesS
    noTree'  <- updateTree (depth + 1) minSamples  noTree  noS
    return $ Branch yesTree' noTree' bestF bestC



-- -----------------------------------------------------------------------------
-- Sample drawing

samplesDrawTrees :: Samples -> RandomForestM [Samples]
samplesDrawTrees smpls = do
  (nT, sF) <- lift $ asksFCfg (nTrees &&& sampleFreq)

  let trs = repeat $ replicateM sF (randomRM (0, nT - 1))
  sampleTrees <- zipWithM (\ s t -> t >>= \ t' -> return (s, t')) smpls trs

  return $ insertAtIndices nT sampleTrees

treesDrawSamples :: Samples -> RandomForestM [Samples]
treesDrawSamples smpls = do
  (nT, sF) <- lift $ asksFCfg (nTrees &&& sampleFreq)

  let samplesPerTree = fI sF * fI (length smpls) / fI nT :: Double
  replicateM nT (sampleWithReplacement (round samplesPerTree) smpls)

  where fI = fromIntegral


-- -----------------------------------------------------------------------------
-- Random Forest

initialiseForest :: LearnData -> Map String String -> (Forest, PredictorCfg Forest)
initialiseForest _ kvs =
  let conf = readConf confActions defaultConf kvs
      g = mkStdGen (seed conf)
      gs = splitMany g
      frst = Forest { trees = zip (replicate (nTrees conf) emptyTree) gs
                    , rng = g } in
  (frst, ForestCfg conf)

-- | Determine which trees should learn which samples, return updated trees.
learnForest :: Forest -> Samples -> ForestM Forest
learnForest frst smpls = do
  sS <- asksFCfg sampleStrat
  (treeSmpls, g') <- runRandomStateT (sampleFunction sS smpls) (rng frst)

  trees' <- parMapReader (uncurry updateTree') $ zip (trees frst) treeSmpls
  return Forest {trees = trees', rng = g'}

  where
    sampleFunction TreesDrawSamples = treesDrawSamples
    sampleFunction SamplesDrawTrees = samplesDrawTrees

-- | Query all trees separately, then combine predictions.
predictForest :: Forest -> Features -> ForestM Labels
predictForest frst ftrs = do
  nP <- asksLData nPredictions
  nT <- asksFCfg nTrees
  cM <- asksFCfg combineMean

  p  <- mapM (query ftrs . fst) (trees frst)
  return $ combine cM nP (replicate nT (1 / fromIntegral nT)) p


-- -----------------------------------------------------------------------------
-- Tree loading/saving

treeToDot :: Tree -> ForestM ByteString
treeToDot tree =
  printTree "t" tree
  >>= \ l -> return $ BC.unlines $ prologue ++ l ++ epilogue where

  prologue = ["digraph tree {", "  node [shape=box]"]
  epilogue = ["}"]

  link from to = BC.unwords [" ", from, "->", to]

  printKVs :: [(ByteString, ByteString)] -> ByteString
  printKVs = B.intercalate ", " . map (\ (k, v) -> B.concat [k, " = ", v])

  quote :: ByteString -> ByteString
  quote s = B.concat ["\"", s, "\""]

  printNode
    :: ByteString -> ByteString -> [(ByteString, ByteString)] -> ByteString
  printNode pos label attrs = B.concat ["  ", pos, " [", kvs, "]"]
    where kvs = printKVs $ ("label", quote label) : attrs

  printColorNode :: ByteString -> ByteString -> Double -> ByteString
  printColorNode pos label weight =
    printNode pos label [("color", cstr), ("style", "filled")]
    where cstr = quote $ B.concat ["0.8 ", BC.pack $ show weight, " 1.0"]

  printLabel :: ByteString -> Int -> ByteString -> [ByteString]
  printLabel basePos i l = [link basePos pos, printNode pos l []]
    where pos = B.append basePos (BC.pack $ show i)

  printTree :: ByteString -> Tree -> ForestM [ByteString]
  printTree pos (Leaf smpls) =
    liftM (concat . zipWith (printLabel pos) [0 ..]) (mapM showSampleM smpls)
  printTree pos (Branch l r f c) = do
    fS <- showFeatureM f
    lT <- printTree ypos l
    rT <- printTree npos r
    return $ [printColorNode pos fS c, link pos ypos, link pos npos] ++
             lT ++ rT
    where ypos = BC.snoc pos 'y'
          npos = BC.snoc pos 'n'

dotToTree :: [(ByteString, ByteString)] -> ForestM Tree
dotToTree allLines@((pos, lbl):xs) =
  if not (B.null pos) && isDigit (BC.head pos)
  then do
    lM <- asksLData labelBM
    return $ Leaf $ map (SampleID . IBM.findInt lM . snd) allLines
  else do
    fM <- asksLData featureBM
    yTree <- dotToTree $ map (first $ B.drop 1) y
    nTree <- dotToTree $ map (first $ B.drop 1) n
    return $ Branch yTree nTree (IBM.findInt fM lbl) 1.0 -- TODO!!
  where (y, n) = span (\ (p, _) -> BC.head p == 'y') xs
dotToTree [] = error "dot file parsing failed: too few lines"


treeFileName :: FilePath -> Int -> FilePath
treeFileName dir i = dir </> "tree" ++ show (i::Int) ++ ".dot"

saveForestGraphs :: Forest -> FilePath -> ForestT IO ()
saveForestGraphs frst dir =
  liftIO (createDirectoryIfMissing True dir) >>
  zipWithM_ save [1 ..] (map fst $ trees frst) where
    save i t = promoteReader (treeToDot t)
                >>= liftIO . B.writeFile (treeFileName dir i)

loadForestGraphs :: Forest -> FilePath -> ForestT IO Forest
loadForestGraphs frst dir =
  mapM load [1 .. length $ trees frst]
  >>= \ trees' -> return frst {trees = zip trees' $ splitMany (rng frst)} where
    load i = liftIO (I.loadDotFile $ treeFileName dir i)
             >>= promoteReader . dotToTree
