{-# LANGUAGE TypeFamilies #-}

module KNN (
  initialiseKNN
, knn
) where

import           Control.Arrow ((&&&), first)
import qualified Data.IntMap.Strict as IM
import qualified Data.IntSet as IS
import qualified Data.List as L
import           Data.Ord (comparing)

import           Predictor
import           Types
import           Utils

data KNN = KNN
  { samplesSeen :: Samples
  , fFreq       :: FeatureFreq
  , fWeights    :: FWeights
  }

instance Predictor KNN where
  data PredictorCfg KNN = KNNCfg KNNConfig

  learn      = learnKNN
  predict    = predictKNN
  saveState  = undefined
  loadState  = undefined

type KNNM   = PredictorM   KNN

type Rank = (Int, Double)
type SWeight = (SampleID, Double)
type FWeights = FeatureMap Double
type Predictions = LabelMap [Rank]
type PredictionsN = (Int, Predictions)


data KNNConfig = KNNConfig {
  depWeight :: Double
, idf     :: Bool
}

defaultConf :: KNNConfig
defaultConf = KNNConfig {
  depWeight = 2.7
, idf       = True
}

confActions :: [(String, KNNConfig -> String -> KNNConfig)]
confActions =
  [ ("depWeight", \ c v -> c {depWeight = read v})
  , ("idf"      , \ c v -> c {idf       = read v})
  ]

asksCfg :: (KNNConfig -> a) -> KNNM a
asksCfg f = askConfig >>= \ (KNNCfg c) -> return $ f c


initialiseKNN :: LearnData -> Configuration -> (KNN, PredictorCfg KNN)
initialiseKNN _ kvs =
  let conf = readConf confActions defaultConf kvs in
  (KNN {samplesSeen = [], fFreq = IM.empty, fWeights = IM.empty}, KNNCfg conf)

learnKNN :: KNN -> Samples -> KNNM KNN
learnKNN (KNN sS fF fW) smpls =
  asksLData sampleFeatures >>= \ fT ->
  asksCfg idf >>= \ iDF ->
  let sS' = sS ++ smpls
      fF' = featureFreq fT fF smpls
      fW' = if iDF then IM.map (** 6) $ featureIDFs (length sS') fF' else fW
  in return KNN
  { samplesSeen = sS'
  , fFreq       = fF'
  , fWeights    = fW'
  }

predictKNN :: KNN -> Features -> KNNM Labels
predictKNN (KNN sS _ fW) ftrs = do
  (lT, fT) <- asksLData (sampleLabels &&& sampleFeatures)
  nP <- asksLData nPredictions
  dW <- asksCfg depWeight

  return $ knn lT fT nP dW ftrs fW $ zip sS $ repeat 1


-- | For each (sample, weight) in the list, determine the correlation (inverse
-- distance) of sample with features by multiplying sample weight with
-- sum of weights of common features.
-- For features without feature weight, use weight of 1.
-- Output (sample, correlation) list sorted by decreasing correlation.
correlate :: SampleFeatures -> Features -> FWeights -> [SWeight] -> [SWeight]
correlate fT ftrs fWs = sort . foldr (uncurry insert) IM.empty where
  insert (SampleID s) w = IM.insertWith (+) s (w * score (SampleID s))
  score = sum . map weight . IS.toList . common
  weight f = IM.findWithDefault 1 f fWs
  common = IS.intersection ftrsS . IS.fromList . getSampleFeatures fT
  sort = L.sortBy (flip $ comparing snd) . map (first SampleID) . IM.toList
  ftrsS = IS.fromList ftrs

-- | Predict labels from sorted (sample, correlation) list, stopping just
-- before the maximal number of predictions is exceeded.
predictLabels :: SampleLabels -> Double -> Int -> [SWeight] -> Predictions
predictLabels lT lW maxPreds =
  snd . last . takeWhile p . scanl (addSample lT lW) (0, IM.empty) . zip [1 ..]
  where p (nPreds, _) = nPreds <= maxPreds

-- | Add sample and its labels (latter with label weight) to predictions.
addSample :: SampleLabels -> Double -> PredictionsN -> (Int, SWeight) -> PredictionsN
addSample lT lW preds (k, (smpl, smplCorr)) =
  foldr (addLabel (k, lblCorr)) (addLabel (k, smplCorr) smplLbl preds) lbls
  where smplLbl = labelOfSample smpl
        lbls = getSampleLabels lT smpl
        lblCorr = lW * smplCorr / log (fromIntegral (length lbls) + 1)

-- | Add label with rank to predictions.
addLabel :: Rank -> LabelID -> PredictionsN -> PredictionsN
addLabel rank lbl (nPreds, preds) = (nPreds', preds') where
  (oldR, preds') = IM.insertLookupWithKey (\ _ _ -> (rank :)) lbl [rank] preds
  -- | if for the label no ranks did exist before, increase nPreds
  nPreds' = maybe (nPreds + 1) (const nPreds) oldR

-- | Given a map from labels to their associated ranks, return labels sorted
-- by decreasing relevance.
sortLabels :: Predictions -> Labels
sortLabels = map fst . L.sortBy (flip $ comparing snd) . IM.toList . IM.map rank
  --map fst . L.sortBy (comparing snd) . IM.toList . IM.map rank
  where rank = sum . map (\ (k, corr) -> corr / log (fromIntegral k + 1))
  {-where rank xs = age (fst (head xs)) + sum (map snd xs)
        age k = 500000000 - 100 * fromIntegral k-}


-- | Run adaptive k-NN on weighted list of samples and list of features.
-- The resulting set of labels is sorted by decreasing relevance and contains
-- as many as possible, but less or equal than `maxPreds` labels.
-- The input samples are also output, but they are weighted differently than
-- their labels, because the latter are first weighted with `lblW`.
knn :: SampleLabels -> SampleFeatures -> Int -> Double -> Features ->
  FWeights -> [SWeight] -> Labels
knn lT fT maxPreds lblW ftrs fWs =
  sortLabels . predictLabels lT lblW maxPreds . correlate fT ftrs fWs
