{-# LANGUAGE TypeFamilies #-}

module NaiveBayes (initialiseNaiveBayes) where

import           Control.Arrow ((&&&), second)
import           Data.Ord (comparing)
import           Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import qualified Data.IntMap.Strict as IM
import           Data.List (sortBy)
import           Data.Monoid (Monoid, mappend, mempty)

import           Predictor
import           Types
import           Utils

data NaiveBayes = NaiveBayes
  { lStats   :: LabelsStats
  , fFreq    :: FeatureFreq
  , fWeights :: FeatureWeights
  }

instance Predictor NaiveBayes where
  data PredictorCfg NaiveBayes = NaiveBayesConfig

  learn      = learnNaiveBayes
  predict    = predictNaiveBayes
  saveState  = undefined
  loadState  = undefined


type NaiveBayesCfg = PredictorCfg NaiveBayes
type NaiveBayesM   = PredictorM   NaiveBayes

type FeatureSet = IntSet
type FeatureWeights = FeatureMap Double


-- -----------------------------------------------------------------------------
-- Statistics for a single label

-- | Statistics for a single label comprise:
-- * in how many samples the label appeared, and
-- * the multiset of those features that appeared in those samples
data LabelStats = LabelStats Int FeatureFreq

instance Monoid LabelStats where
  mempty = LabelStats 0 IM.empty
  mappend (LabelStats lF1 fF1) (LabelStats lF2 fF2) =
    LabelStats (lF1 + lF2) (IM.unionWith (+) fF1 fF2)

-- | Create stats for label that appeared n times together with given features.
initStats :: Int -> Features -> LabelStats
initStats w ftrs = LabelStats w (IM.fromList $ zip ftrs $ repeat w)

-- | Correlation of a given label with a weighted feature set.
score :: FeatureSet -> FeatureWeights -> LabelStats -> Double
score ftrs weights (LabelStats revDeps fF) = 30 * log (revDepsN)
  +       sum (mapIM (\ (f, x) -> weight f * log (5 *    num x  / revDepsN)) s1)
  + 0.2 * sum (mapIM (\ (f, x) -> weight f * log (1 + (1-num x) / revDepsN)) s2)
  -  18 * sum (mapIS weight s3)

  where
    num = fromIntegral
    revDepsN = num revDeps

    weight ftr = IM.findWithDefault 1 ftr weights

    -- s1 contains features appearing both in label stats and query features,
    -- s2 contains features only in label stats, and
    -- s3 contains features only in query features
    (s1, s2) = IM.partitionWithKey (\ f _ -> f `IS.member` ftrs) fF
    s3 = ftrs `IS.difference` IM.keysSet s1


-- -----------------------------------------------------------------------------
-- Statistics for multiple labels

type LabelsStats = LabelMap LabelStats

-- | Create stats for label and add them to the statistics for all labels.
updateStats :: Int -> Features -> LabelID -> LabelsStats -> LabelsStats
updateStats w ftrs lbl = IM.insertWith mappend lbl $ initStats w ftrs

-- | Learn that the sample labels appeared together with the sample features,
-- as well as that the sample itself appeared with the sample features (with
-- a higher importance).
learnSample :: Labels -> Features -> SampleID -> LabelsStats -> LabelsStats
learnSample lbls ftrs smpl = updateSample . updateLabels where
  updateLabels stats = foldr (updateStats 1 ftrs) stats lbls
  updateSample = updateStats 1000 ftrs (labelOfSample smpl)


-- -----------------------------------------------------------------------------
-- Naive Bayes

initialiseNaiveBayes :: LearnData -> Configuration -> (NaiveBayes, NaiveBayesCfg)
initialiseNaiveBayes _ _ =
  (NaiveBayes {lStats = IM.empty, fFreq = IM.empty, fWeights = IM.empty},
   NaiveBayesConfig)

learnNaiveBayes :: NaiveBayes -> Samples -> NaiveBayesM NaiveBayes
learnNaiveBayes (NaiveBayes lS fF _) smpls =
  asksLData (sampleLabels &&& sampleFeatures) >>= \ (sL, sF) ->
  let f smpl = learnSample (getSampleLabels   sL smpl)
                           (getSampleFeatures sF smpl) smpl
      lS' = foldr f lS smpls
      fF' = featureFreq sF fF smpls

  in return NaiveBayes
  { lStats   = lS'
  , fFreq    = fF'
  , fWeights = featureIDFs (IM.size lS') fF'
  }

predictNaiveBayes :: NaiveBayes -> Features -> NaiveBayesM Labels
predictNaiveBayes (NaiveBayes lS _ fW) ftrs = return $
  map fst $ sortBy (flip $ comparing snd) $ mapIM (second $ score ftrS fW) lS
  where ftrS = IS.fromList ftrs
