{-# LANGUAGE OverloadedStrings #-}

module Parser
( parseFeaturesLine
, parseLabelsLine
, parsePredictorOptions
, parseStringList
, parseDotLine
) where

{-
Lessons learnt from writing the tokenizer:
- First attempt using String and read was horribly slow --- about 30s for 20M.
- Second attempt using Attoparsec.ByteString.Char8 was already relatively
  fast, but took lots of memory.
- Third attempt improved second attempt by using sepBy' instead of sepBy,
  thanks to Andreas Kochesser.
- Fourth attempt splits lines manually (using ByteString.lines) before
  handing them to the parser, resulting in still better speeds. Before that,
  the parser was responsible for detecting new lines. As a bonus, we can now
  evaluate the result in a lazy manner, e.g. get the number of theorems
  in the file by doing a length, which will terminate very quickly.
- Fifth attempt is a hand-rolled parser using ByteString, which has speeds
  quite comparable to the parser written in C.
-}

import           Control.Arrow (second)
import           Data.Attoparsec.ByteString (Parser)
import qualified Data.Attoparsec.ByteString.Char8 as AC
import           Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Char8 as BC
import           Data.List.Split (splitWhen)
import           Data.Map (Map)
import qualified Data.Map as Map


type BBList = (ByteString, [ByteString])


-- adapted from Haskell's documentation for Data.ByteString such that
-- reading an empty string as input will not result in an empty string,
-- but an empty list (important for label lines without any labels)
tokenise :: ByteString -> ByteString -> [ByteString]
tokenise x y =
  if B.null h then [] else
    h : if B.null t then [] else tokenise x (B.drop (B.length x) t)
  where (h,t) = B.breakSubstring x y

-- splits a line of the format "part1:part2" into ("part1", "part2")
splitIDLine :: ByteString -> (ByteString, ByteString)
splitIDLine = second B.tail . BC.break (== ':')

-- parse one line of the format: "feat_1", "feat_2", ..., "feat_n"
-- we do not check whether the features are actually delimited by quotes!
-- TODO: be more strict about that in the future ...
parseStringList :: ByteString -> [ByteString]
parseStringList = map removeQuotes . tokenise ", "
  where removeQuotes = B.init . B.tail

-- parse one line of the format: thm_id:"feat_1", "feat_2", ..., "feat_n"
parseFeaturesLine :: ByteString -> BBList
parseFeaturesLine = second parseStringList . splitIDLine

-- parse one line of the format: thm_id:lbl_1 lbl_2 ... lbl_n
parseLabelsLine :: ByteString -> BBList
parseLabelsLine = second (tokenise " ") . splitIDLine

-- parse strings of the format: key=value
parseKeyValueStrings :: [String] -> Map String String
parseKeyValueStrings = foldr handleKV Map.empty where
  handleKV kv m = case splitWhen (== '=') kv of
    [k, v] -> Map.insert k v m
    _ -> error "invalid key/value format"

-- parse string of the format: id:key1=value1:key2=value2:...:key_n=value_n
parsePredictorOptions :: String -> (String, Map String String)
parsePredictorOptions s =
  case splitWhen (== ':') s of
    [] -> error "no predictor given"
    p : args -> (p, parseKeyValueStrings args)

-- parse a quoted string, such as "string"
quotedString :: Parser ByteString
quotedString =
  AC.string "\""
  >> AC.takeWhile (/= '"')
  >>= \ s -> AC.string "\""
  >> return s

-- parse line of the format "  t(lr)*[0-9]* pos [label = \"name\"]" and
-- return ("(lr)*[0-9]*", "name")
dotLineParser :: Parser (ByteString, ByteString)
dotLineParser =
  AC.skipSpace
  >> AC.string "t"
  >> AC.takeWhile (/= ' ')
  >>= \ pos -> AC.string " [label = "
  >> quotedString
  >>= \ lbl -> AC.string "]"

  >> return (pos, lbl)

parseDotLine :: ByteString -> Maybe (ByteString, ByteString)
parseDotLine = AC.maybeResult . AC.parse dotLineParser
