module Utils where

import           Control.Arrow ((&&&), second)
import           Control.DeepSeq (NFData)
import           Control.Monad ((>=>))
import           Control.Monad.IO.Class (MonadIO, liftIO)
import           Control.Monad.Reader (Reader, ask, runReader)
import           Control.Monad.Reader.Class (MonadReader)
import           Control.Parallel.Strategies (parMap, rdeepseq)
import           Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Char8 as BC
import           Data.Foldable (toList)
import           Data.IntMap (IntMap)
import qualified Data.IntMap as IM
import           Data.IntSet (IntSet)
import qualified Data.IntSet as IS
import qualified Data.List as L
import           Data.Map (Map)
import qualified Data.Map as Map
import           Data.Ord (comparing)
import qualified Data.Sequence as Seq

import           Debug.Trace


-- -----------------------------------------------------------------------------
-- Random stuff

-- | Prints given value in a pure computation.
-- Taken from: http://en.wikibooks.org/wiki/Haskell/Debugging
traceThis :: (Show a) => a -> a
traceThis x = trace (show x) x

-- | Find minimal value (in contrast to key) of a Map.
findMinValue :: Ord a => Map k a -> (k, a)
findMinValue = L.minimumBy (comparing snd) . Map.toList

-- | Given a list of elements with indices, create a list which will contain
-- at each index all elements which are associated with that index.
insertAtIndices :: Int -> [(a, [Int])] -> [[a]]
insertAtIndices maxIndex =
  toList . foldl (\ acc (x, is) -> foldr (Seq.adjust (x :)) acc is)
           (Seq.replicate maxIndex [])

-- | Execute a Reader function from a Reader monad in parallel on a list.
parMapReader
  :: (NFData b, MonadReader r m) => (a -> Reader r b) -> [a] -> m [b]
parMapReader f l =
  ask >>= \ r -> return $ parMap rdeepseq (\ x -> runReader (f x) r) l


-- | Execute monadic function on each line of a ByteString.
linesInteractM :: MonadIO m => (ByteString -> m ByteString) -> m ()
linesInteractM f =
  liftIO B.getContents >>= mapM_ (f >=> liftIO . BC.putStrLn) . BC.lines


-- | Given lists [1, 2, 3] and [42, 1, 9], this will return [43, 3, 12].
sumLists :: Num a => [[a]] -> [a]
sumLists = foldl1 (zipWith (+))

-- | Map on list representation of IntSet.
mapIS :: (Int -> b) -> IntSet -> [b]
mapIS f = map f . IS.toList

-- | Map on list representation of IntMap.
mapIM :: ((Int, a) -> b) -> IntMap a -> [b]
mapIM f = map f . IM.toList

-- | Map on IntMap elements, and return list sorted by output value.
sortMapIM :: Ord b => (a -> b) -> IntMap a -> [(Int, b)]
sortMapIM f = L.sortBy (comparing snd) . map (second f) . IM.toList


-- -----------------------------------------------------------------------------
-- Pairs

-- | Execute function on both values of a pair.
mapPair :: (a -> b) -> (a, a) -> (b, b)
mapPair f (x, y) = (f x, f y)

-- | Apply a monadic function to the first element of a pair.
firstM :: Monad m => (t -> m t1) -> (t, t2) -> m (t1, t2)
firstM f (a, b) = f a >>= \ a' -> return (a', b)

-- | Evaluate monadic values in pair and put result into monad.
liftPair :: Monad m => (m a, m b) -> m (a, b) 
liftPair (x, y) = x >>= \ x' -> y >>= \ y' -> return (x', y')

-- | Evaluate second part of a pair and put whole pair into monad.
liftSnd :: Functor m => (a, m b) -> m (a, b)
liftSnd (x, y) = fmap (\ y' -> (x, y')) y

distributeFst :: [(a, [b])] -> [(a, b)]
distributeFst = concatMap (\ (x, ys) -> map (\ y -> (x, y)) ys)

intMapFromPairs :: [(Int, a)] -> IntMap [a]
intMapFromPairs = foldr (\ (k, x) -> IM.insertWith (++) k [x]) IM.empty


-- -----------------------------------------------------------------------------
-- Statistics

-- | For each frequency, store number of elements occurring with that frequency.
newtype Histogram = Histogram [(Int, Int)]

instance Show Histogram where
  show (Histogram h) = unlines $ map (\ (o, f) -> show o ++ " " ++ show f) h


-- | Compute arithmetic mean.
average :: Fractional a => [a] -> a
average xs = sum xs / fromIntegral (length xs)

-- | Compute for each list element how often it appears in the list.
frequency :: Ord a => [a] -> [(a, Int)]
frequency = map (head &&& length) . L.group . L.sort

-- | Compute histogram.
histogram :: Ord a => [a] -> Histogram
histogram = Histogram . frequency . map snd . frequency


-- -----------------------------------------------------------------------------
-- Configuration building

updateConf :: (Ord k) => [(k, c -> v -> c)] -> k -> v -> c -> c
updateConf cfs k v conf = case lookup k cfs of
  Nothing -> error "configuration option not found"
  Just f -> f conf v

readConf :: (Ord k) => [(k, c -> v -> c)] -> c -> Map k v -> c
readConf cfs = Map.foldrWithKey (updateConf cfs)
